import copy

import numpy
import pandas
import pandas as pd
import pyod
import sklearn
from logitboost import LogitBoost
from pyod.models.abod import ABOD
from pyod.models.base import BaseDetector
from pyod.models.cblof import CBLOF
from pyod.models.cof import COF
from pyod.models.copod import COPOD
from pyod.models.ecod import ECOD
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.inne import INNE
from pyod.models.knn import KNN
from pyod.models.loda import LODA
from pyod.models.lof import LOF
from pyod.models.mcd import MCD
from pyod.models.ocsvm import OCSVM
from pyod.models.pca import PCA
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier, BaggingClassifier, GradientBoostingClassifier
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, LabelEncoder
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, check_array
from xgboost import XGBClassifier

from src.utils.general_utils import current_ms
# ---------------------------------- SUPPORT METHODS ------------------------------------
from src.metrics.EnsembleMetric import get_default


def get_classifier_name(clf):
    if isinstance(clf, Classifier):
        return clf.classifier_name()
    else:
        return clf.__class__.__name__


def get_feature_importance(clf):
    if isinstance(clf, Classifier):
        return clf.feature_importances()
    else:
        return clf.feature_importances_


def auto_bag_rate(n_features):
    """
    Method used to automatically devise a rate to include features in bagging
    :param n_features: number of features in the dataset
    :return: the rate of features to bag
    """
    if n_features < 20:
        bag_rate = 0.8
    elif n_features < 50:
        bag_rate = 0.7
    elif n_features < 100:
        bag_rate = 0.6
    else:
        bag_rate = 0.5
    return bag_rate


def predict_uns_proba(uns_clf, test_features):
    proba = uns_clf.predict_proba(test_features)
    pred = numpy.argmax(proba, axis=1)
    for i in range(len(pred)):
        min_p = min(proba[i])
        max_p = max(proba[i])
        proba[i][pred[i]] = max_p
        proba[i][1 - pred[i]] = min_p
    return proba


class Classifier(BaseEstimator, ClassifierMixin):
    """
    Basic Abstract Class for Classifiers.
    Abstract methods are only the classifier_name, with many degrees of freedom in implementing them.
    Wraps implementations from different frameworks (if needed), sklearn and many deep learning utilities
    """

    def __init__(self, clf):
        """
        Constructor of a generic Classifier
        :param clf: algorithm to be used as Classifier
        """
        self.clf = copy.deepcopy(clf) if clf is not None else None
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        if y is not None:
            self.classes_ = unique_labels(y)
        else:
            self.classes_ = [0, 1]
        # self.X_ = X
        # self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def predict(self, X):
        """
        Method to compute predict of a classifier
        :return: array of predicted class
        """
        probas = self.predict_proba(X)
        if self.is_unsupervised():
            return numpy.argmax(probas, axis=1)
        else:
            return self.classes_[numpy.argmax(probas, axis=1)]

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        return self.clf.predict_proba(X)

    def predict_confidence(self, X):
        """
        Method to compute confidence in the predicted class
        :return: max probability as default
        """
        probas = self.predict_proba(X)
        return numpy.max(probas, axis=1)

    def compute_feature_importances(self):
        """
        Outputs feature ranking in building a Classifier
        :return: ndarray containing feature ranks
        """
        if hasattr(self.clf, 'feature_importances_'):
            return self.clf.feature_importances_
        elif hasattr(self.clf, 'coef_'):
            return numpy.sum(numpy.absolute(self.clf.coef_), axis=0)
        return []

    def is_unsupervised(self):
        """
        true if the classifier is unsupervised
        :return: boolean
        """
        return hasattr(self, 'classes_') and numpy.array_equal(self.classes_, [0, 1])

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__

    def get_params(self, deep=True):
        """
        Compliance with scikit-learn
        :param deep: see scikit-learn
        :return: see scikit-learn
        """
        return {'clf': self.clf}

    def get_diversity(self, X, y, metrics=None):
        """
        Returns diversity metrics. Works only with ensembles.
        :param metrics: name of the metrics to output (list of Metric objects)
        :param X: test set
        :param y: labels of the test set
        :return: diversity metrics
        """
        X = check_array(X)
        predictions = []
        check_is_fitted(self)
        if hasattr(self, "estimators_"):
            # If it is an ensemble and if it is trained
            for baselearner in self.estimators_:
                predictions.append(baselearner.predict(X))
            predictions = numpy.column_stack(predictions)
        elif self.clf is not None:
            # If it wraps an ensemble
            check_is_fitted(self.clf)
            if hasattr(self.clf, "estimators_"):
                # If it is an ensemble and if it is trained
                for baselearner in self.clf.estimators_:
                    predictions.append(baselearner.predict(X))
                predictions = numpy.column_stack(predictions)
        if predictions is not None and len(predictions) > 0:
            # Compute metrics
            metric_scores = {}
            if metrics is None or not isinstance(metrics, list):
                metrics = get_default()
            for metric in metrics:
                metric_scores[metric.get_name()] = metric.compute_diversity(predictions, y)
            return metric_scores
        else:
            # If it is not an ensemble
            return {}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self.clf, parameter, value)
        return self


class UnsupervisedClassifier(Classifier, BaseDetector):
    """
    Wrapper for unsupervised classifiers belonging to the library PYOD
    """

    def __init__(self, clf):
        """
        Constructor of a generic UnsupervisedClassifier. Assumes that clf is an algorithm from pyod
        :param clf: pyod algorithm to be used as Classifier
        """
        self.clf = clf
        self.contamination = clf.contamination
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):
        # Store the classes seen during fit + other data
        self.classes_ = [0, 1]
        self.X_ = X
        self.y_ = None

        # Train clf
        self.clf.fit(X)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def decision_function(self, X):
        """
        pyod function to override. Calls the wrapped classifier.
        :param X: test set
        :return: decision function
        """
        return self.clf.decision_function(X)

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes.
        It has to e overridden since PYOD's implementation of predict_proba is wrong
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        pred_score = self.decision_function(X)
        probs = numpy.zeros((X.shape[0], 2))
        if isinstance(self.contamination, (float, int)):
            pred_thr = pred_score - self.clf.threshold_
        min_pt = min(pred_thr)
        max_pt = max(pred_thr)
        anomaly = pred_thr > 0
        cont = numpy.asarray([pred_thr[i] / max_pt if anomaly[i] else (pred_thr[i] / min_pt if min_pt != 0 else 0.2)
                              for i in range(0, len(pred_thr))])
        probs[:, 0] = 0.5 + cont / 2
        probs[:, 1] = 1 - probs[:, 0]
        probs[anomaly, 0], probs[anomaly, 1] = probs[anomaly, 1], probs[anomaly, 0]
        return probs

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__


class XGB(Classifier):
    """
    Wrapper for the xgboost.XGBClassifier algorithm
    Can be used as reference to see what's needed to extend the Classifier class
    """

    def __init__(self, n_estimators=100):
        Classifier.__init__(self, XGBClassifier(n_estimators=n_estimators))
        self.l_encoder = None

    def fit(self, X, y=None):
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        self.classes_ = unique_labels(y)
        self.l_encoder = LabelEncoder()
        y = self.l_encoder.fit_transform(y)

        # self.X_ = X
        # self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def classifier_name(self):
        return "XGBClassifier"


import copy
import multiprocessing
import random
from multiprocessing import Process

import numpy
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, check_array

from src.classifiers.Classifier import Classifier
from src.metrics.EnsembleMetric import get_default


class ConfidenceBagging(Classifier):
    """
    Class for creating bagging ensembles
    """

    def __init__(self, clf, n_base: int = 10, max_features: float = 0.7, sampling_ratio: float = 0.7,
                 perc_decisors: float = None, n_decisors: int = None, weighted: bool = False):
        """
        Constructor
        :param clf: the algorithm to be used for creating base learners
        :param n_base: number of base learners (= size of the ensemble)
        :param max_features: percentage of features to be used at each iteration
        :param sampling_ratio: percentage of the dataset to be used at each iteration
        :param perc_decisors: percentage of base learners to be used for prediction
        :param n_decisors: number of base learners to be used for prediction
        :param weighted: True if prediction has to be computed as a weighted sum of probabilities
        """
        super().__init__(clf)
        self.weighted = weighted
        if n_base > 1:
            self.n_base = n_base
        else:
            print("Ensembles have to be at least 2")
            self.n_base = 10

        self.max_features = max_features if max_features is not None and 0 < max_features <= 1 else 0.7

        self.sampling_ratio = sampling_ratio if sampling_ratio is not None and 0 < sampling_ratio <= 1 else 0.7

        if perc_decisors is not None and 0 < perc_decisors <= 1:
            if n_decisors is not None and 0 < n_decisors <= self.n_base:
                print('Both perc_decisors and n_decisors are specified, prioritizing perc_decisors')
            self.n_decisors = int(self.n_base*perc_decisors) if int(self.n_base*perc_decisors) > 0 else 1
        elif n_decisors is not None and 0 < n_decisors <= self.n_base:
            self.n_decisors = n_decisors
        else:
            self.n_decisors = 1 + int(self.n_base / 2)
        self.estimators_ = []
        self.feature_sets = []

    def fit_base(self, X, y, n_samples, f_bag, f_list, c_list, lock):
        # Draw samples
        features = random.sample(range(X.shape[1]), f_bag)
        features.sort()
        sample_x, sample_y = self.draw_samples(X, y, n_samples)
        sample_x = sample_x[:, features]
        if len(features) == 1:
            sample_x = sample_x.reshape(-1, 1)

        # Train learner
        learner = copy.deepcopy(self.clf)
        learner.fit(sample_x, sample_y)
        if hasattr(learner, "X_"):
            learner.X_ = None
        if hasattr(learner, "y_"):
            learner.y_ = None

        # Test Learner
        with lock:
            f_list.append(features)
            c_list.append(learner)

    def fit(self, X, y=None):
        train_n = len(X)
        self.classes_ = unique_labels(y) if y is not None else [0, 1]
        bag_features_n = int(X.shape[1]*self.max_features)
        samples_n = int(train_n * self.sampling_ratio)

        # Parallel Training
        f_list = multiprocessing.Manager().list()
        c_list = multiprocessing.Manager().list()
        lock = multiprocessing.Lock()
        train_tasks = [Process(target=self.fit_base,
                               kwargs={'X': X, 'y': y, 'n_samples': samples_n, 'f_bag': bag_features_n,
                                       'f_list': f_list, 'c_list': c_list, 'lock': lock})
                       for _ in range(0, self.n_base)]
        for running_task in train_tasks:
            running_task.start()
        for running_task in train_tasks:
            running_task.join()

        self.feature_sets = [x for x in f_list]
        self.estimators_ = [x for x in c_list]

        # Compliance with SKLEARN and PYOD
        self.X_ = X[[0, 1], :]
        self.y_ = y
        self.feature_importances_ = self.compute_feature_importances()

    def draw_samples(self, X, y, samples_n):
        indexes = numpy.random.choice(X.shape[0], samples_n, replace=False, p=None)
        sample_x = numpy.asarray(X[indexes, :])
        # If data is labeled we also have to refine labels
        if y is not None and hasattr(self, 'classes_') and self.classes_ is not None and len(self.classes_) > 1:
            sample_y = y[indexes]
            sample_labels = unique_labels(sample_y)
            missing_labels = [item for item in self.classes_ if item not in sample_labels]
            # And make sure that there is at least a sample for each class of the problem
            if missing_labels is not None and len(missing_labels) > 0:
                # For each missing class
                for missing_class in missing_labels:
                    miss_class_indexes = numpy.asarray(numpy.where(y == missing_class)[0])
                    new_sampled_index = numpy.random.choice(miss_class_indexes, None, replace=False, p=None)
                    X_missing_class = X[new_sampled_index, :]
                    sample_x = numpy.append(sample_x, [X_missing_class], axis=0)
                    sample_y = numpy.append(sample_y, missing_class)
        else:
            sample_y = None
        return sample_x, sample_y

    def predict_proba(self, X):
        # Scoring probabilities, ends with a
        proba_array = []
        conf_array = []
        for i in range(0, self.n_base):
            predictions = self.estimators_[i].predict_proba(X[:, self.feature_sets[i]])
            proba_array.append(predictions)
            conf_array.append(numpy.max(predictions, axis=1))
        # 3d matrix (clf, row, probability for class)
        proba_array = numpy.asarray(proba_array)
        # 2dim matrix (clf, confidence for row)
        conf_array = numpy.asarray(conf_array)

        # Choosing the most confident self.n_decisors to compute final probabilities
        proba = numpy.zeros(proba_array[0].shape)
        if self.weighted:
            for i in range(0, X.shape[0]):
                proba[i] = numpy.sum(proba_array[:, i, :].T * conf_array[:, i], axis=1) / numpy.sum(conf_array[:, i])
        else:
            conf_array = conf_array.transpose()
            all_conf = -numpy.sort(-conf_array, axis=1)
            conf_thrs = all_conf[:, self.n_decisors-1]
            for i in range(0, X.shape[0]):
                proba[i] = numpy.average(proba_array[numpy.where(conf_array[i] >= conf_thrs[i]), i, :], axis=1)

        # Final averaged Result
        return proba

    def get_diversity(self, X, y, metrics=None):
        """
        Returns diversity metrics. Works only with ensembles.
        :param metrics: name of the metrics to output (list of Metric objects)
        :param X: test set
        :param y: labels of the test set
        :return: diversity metrics
        """
        X = check_array(X)
        predictions = []
        check_is_fitted(self)
        if hasattr(self, "estimators_"):
            # If it is an ensemble and if it is trained
            for i in range(0, self.n_base):
                predictions.append(self.estimators_[i].predict(X[:, self.feature_sets[i]]))
            predictions = numpy.column_stack(predictions)

        if predictions is not None and len(predictions) > 0:
            # Compute metrics
            metric_scores = {}
            if metrics is None or not isinstance(metrics, list):
                metrics = get_default()
            for metric in metrics:
                metric_scores[metric.get_name()] = metric.compute_diversity(predictions, y)
            return metric_scores
        else:
            # If it is not an ensemble
            return {}

    def classifier_name(self):
        """
        Gets classifier name as string
        :return: the classifier name
        """
        clf_name = self.clf.classifier_name() if isinstance(self.clf, Classifier) else self.clf.__class__.__name__
        if clf_name == 'Pipeline':
            keys = list(self.clf.named_steps.keys())
            clf_name = str(keys) if len(keys) != 2 else str(keys[1]).upper()
        if self.weighted:
            return "ConfidenceBaggerWeighted(" + str(clf_name) + "-" + \
                   str(self.n_base) + "-" + str(self.n_decisors) + "-" + \
                   str(self.max_features) + "-" + str(self.sampling_ratio) + ")"
        else:
            return "ConfidenceBagger(" + str(clf_name) + "-" + \
               str(self.n_base) + "-" + str(self.n_decisors) + "-" + \
               str(self.max_features) + "-" + str(self.sampling_ratio) + ")"



# Support libs
import random

import numpy
import pandas
import sklearn.metrics as metrics
import sklearn.model_selection as ms
# Used to save a classifier and measure its size in KB
from pyod.models.pca import PCA

# Name of the folder in which look for tabular (CSV) datasets
from src.classifiers.Classifier import UnsupervisedClassifier
from src.classifiers.ConfidenceBoosting import ConfidenceBoosting

# Scikit-Learn algorithms
# The PYOD library contains implementations of unsupervised classifiers.
# Works only with anomaly detection (no multi-class)
# ------- GLOBAL VARS -----------

CSV_FILE = "sample_data/sample_data_arancino.csv"
# Name of the column that contains the label in the tabular (CSV) dataset
LABEL_NAME = 'multilabel'
# Name of the 'normal' class in datasets. This will be used only for binary classification (anomaly detection)
NORMAL_TAG = 'normal'
# Percantage of test data wrt train data
TT_SPLIT = 0.5
# True if debug information needs to be shown
VERBOSE = True

# Set random seed for reproducibility
random.seed(42)
numpy.random.seed(42)


# ----------------------- MAIN ROUTINE ---------------------
# This files shows an example of confidence ensembles for unsupervised learning.
# Can be used as is, just change the details to load the dataset

if __name__ == '__main__':

    # if file is a CSV, it is assumed to be a dataset to be processed
    df = pandas.read_csv(CSV_FILE, sep=",")
    if VERBOSE:
        print("\n------------ DATASET INFO -----------------")
        print("Data Points in Dataset '%s': %d" % (CSV_FILE, len(df.index)))
        print("Features in Dataset: " + str(len(df.columns)))

    # Set up train test split excluding categorical values that some algorithms cannot handle
    # 1-Hot-Encoding or other approaches may be used instead of removing
    y = numpy.where(df[LABEL_NAME] == "normal", 0, 1)
    an_perc = numpy.average(y)
    x_train, x_test, y_train, y_test = ms.train_test_split(df.drop(columns=[LABEL_NAME]).to_numpy(), y,
                                                           test_size=TT_SPLIT, shuffle=True)
    # Creating classifiers
    # Note that unsupervised classifiers in PYOD require a contamination
    classifier = PCA(contamination=an_perc)
    cb_clf = ConfidenceBoosting(clf=classifier, n_base=10, learning_rate=2,
                                sampling_ratio=0.5, conf_thr=0.8)

    # Exercising Classifier
    classifier.fit(x_train)
    clf_pred = classifier.predict(x_test)

    # Exercising ConfBoost
    cb_clf.fit(x_train)
    cb_pred = cb_clf.predict(x_test)

    print('PCA has accuracy of %.3f, whereas the ConfBoost(PCA) has accuracy of %.3f' %
          (metrics.accuracy_score(y_test, clf_pred), metrics.accuracy_score(y_test, cb_pred)))



import copy

import numpy
from sklearn.utils.multiclass import unique_labels

from src.classifiers.Classifier import Classifier


class ConfidenceBoosting(Classifier):
    """
    Class for creating Confidence Boosting ensembles
    """

    def __init__(self, clf, n_base: int = 10, learning_rate: float = None,
                 sampling_ratio: float = 0.5, conf_thr: float = 0.8, weighted: bool = False):
        """
        Constructor
        :param clf: the algorithm to be used for creating base learners
        :param n_base: number of base learners (= size of the ensemble)
        :param learning_rate: learning rate for updating dataset weights
        :param sampling_ratio: percentage of the dataset to be used at each iteration
        :param conf_thr: threshold of acceptance for confidence scores. Lower confidence means untrustable result
        :param weighted: True if prediction has to be computed as a weighted sum of probabilities
        """
        super().__init__(clf)
        self.weighted = weighted
        self.proba_thr = None
        self.conf_thr = conf_thr
        self.contamination = clf.contamination if hasattr(clf, 'contamination') else None
        if n_base > 1:
            self.n_base = n_base
        else:
            print("Ensembles have to be at least 2")
            self.n_base = 10
        if learning_rate is not None:
            self.learning_rate = learning_rate
        else:
            self.learning_rate = 2
        if sampling_ratio is not None:
            self.sampling_ratio = sampling_ratio
        else:
            self.sampling_ratio = 1 / n_base ** (1 / 2)
        self.estimators_ = []

    def fit(self, X, y=None):
        """
        Training function for the confidence boosting ensemble
        :param y: labels of the train set (optional, not required for unsupervised learning)
        :param X: train set
        """
        train_n = len(X)
        self.classes_ = unique_labels(y) if y is not None else [0, 1]
        samples_n = int(train_n * self.sampling_ratio)
        weights = numpy.full(train_n, 1 / train_n)
        for learner_index in range(0, self.n_base):
            # Draw samples
            sample_x, sample_y = self.draw_samples(X, y, weights, samples_n)
            # Train learner
            learner = copy.deepcopy(self.clf)
            learner.fit(sample_x, sample_y)
            if hasattr(learner, "X_"):
                learner.X_ = None
            if hasattr(learner, "y_"):
                learner.y_ = None
            # Test Learner
            y_proba = learner.predict_proba(X)
            y_conf = numpy.max(y_proba, axis=1)
            p_thr = self.define_conf_thr(target=self.conf_thr, confs=y_conf)
            self.estimators_.append(learner)
            # Update Weights
            update_flag = numpy.where(y_conf >= p_thr, 0, 1)
            weights = weights * (1 + self.learning_rate * update_flag)
            weights = weights / sum(weights)
        self.proba_thr = self.define_proba_thr(target=self.contamination,
                                               probs=self.predict_proba(X)) \
            if self.contamination is not None else 0.5

        # Compliance with SKLEARN and PYOD
        self.X_ = X[[0, 1], :]
        self.y_ = y
        self.feature_importances_ = self.compute_feature_importances()

    def draw_samples(self, X, y, weights, samples_n):
        indexes = numpy.random.choice(len(weights), samples_n, replace=False, p=weights)
        sample_x = numpy.asarray(X[indexes, :])
        # If data is labeled we also have to refine labels
        if y is not None and hasattr(self, 'classes_') and self.classes_ is not None and len(self.classes_) > 1:
            sample_y = y[indexes]
            sample_labels = unique_labels(sample_y)
            missing_labels = [item for item in self.classes_ if item not in sample_labels]
            # And make sure that there is at least a sample for each class of the problem
            if missing_labels is not None and len(missing_labels) > 0:
                # For each missing class
                for missing_class in missing_labels:
                    miss_class_indexes = numpy.asarray(numpy.where(y == missing_class)[0])
                    new_sampled_index = numpy.random.choice(miss_class_indexes, None, replace=False, p=None)
                    X_missing_class = X[new_sampled_index, :]
                    sample_x = numpy.append(sample_x, [X_missing_class], axis=0)
                    sample_y = numpy.append(sample_y, missing_class)
        else:
            sample_y = None
        return sample_x, sample_y

    def define_conf_thr(self, confs, target=None, delta=0.01):
        target_thr = target
        left_bound = min(confs)
        right_bound = max(confs)
        c_thr = (right_bound + left_bound) / 2
        a = numpy.average(confs < 0.6)
        b = numpy.average(confs < 0.9)
        actual_thr = numpy.average(confs < c_thr)
        while abs(actual_thr - target_thr) > delta and abs(right_bound - left_bound) > 0.001:
            if actual_thr < target_thr:
                left_bound = c_thr
                c_thr = (c_thr + right_bound) / 2
            else:
                right_bound = c_thr
                c_thr = (c_thr + left_bound) / 2
            actual_thr = numpy.average(confs < c_thr)
        return c_thr

    def define_proba_thr(self, probs, target=None, delta=0.01):
        """
        Method for finding a confidence threshold based on the expected contamination (iterative)
        :param probs: probabilities to find threshold of
        :param target: the quantity to be used as reference for gettng to the threshold
        :param delta: the tolerance to stop recursion
        :return: a float value to be used as threshold for updating weights in boosting
        """
        target_cont = target
        p_thr = 0.5
        left_bound = 0.5
        right_bound = 1
        actual_cont = numpy.average(probs[:, 0] < p_thr)
        while abs(actual_cont - target_cont) > delta and abs(right_bound - left_bound) > 0.01:
            if actual_cont < target_cont:
                left_bound = p_thr
                p_thr = (p_thr + right_bound) / 2
            else:
                right_bound = p_thr
                p_thr = (p_thr + left_bound) / 2
            actual_cont = numpy.average(probs[:, 0] < p_thr)
        return p_thr

    def predict_proba(self, X):
        """
        Method to compute prediction probabilities (i.e., normalized logits) of a classifier
        :param X: the test set
        :return: array of probabilities for each data point and each class
        """
        proba_array = []
        conf_array = []
        proba = numpy.zeros((X.shape[0], len(self.classes_)))
        for clf in self.estimators_:
            predictions = clf.predict_proba(X)
            proba += predictions
            proba_array.append(predictions)
            conf_array.append(numpy.max(predictions, axis=1))
            # 3d matrix (clf, row, probability for class)
        proba_array = numpy.asarray(proba_array)
        # 2dim matrix (clf, confidence for row)
        conf_array = numpy.asarray(conf_array)

        if self.weighted:
            # Weighting probas using confidence
            proba = numpy.zeros(proba_array[0].shape)
            for i in range(0, X.shape[0]):
                proba[i] = numpy.sum(proba_array[:, i, :].T * conf_array[:, i], axis=1) / numpy.sum(conf_array[:, i])
        else:
            proba = proba / self.n_base
        # Final averaged Result
        return proba

    def predict_confidence(self, X):
        """
        Method to compute the confidence in predictions of a classifier
        :param X: the test set
        :return: array of confidence scores
        """
        conf = numpy.zeros(X.shape[0])
        for clf in self.estimators_:
            if hasattr(clf, 'predict_confidence') and callable(clf.predict_confidence):
                c_conf = clf.predict_confidence(X)
            else:
                y_proba = clf.predict_proba(X)
                c_conf = numpy.max(y_proba, axis=1)
            conf += c_conf
        return conf / self.n_base

    def get_feature_importances(self):
        """
        Placeholder, to be implemented if possible
        :return: feature importances (to be tested)
        """
        fi = []
        for clf in self.estimators_:
            c_fi = clf.get_feature_importances()
            fi.append(c_fi)
        fi = numpy.asarray(fi)
        return numpy.average(fi, axis=1)

    def predict(self, X):
        """
        Method to compute predict of a classifier
        :param X: the test set
        :return: array of predicted class
        """
        proba = self.predict_proba(X)
        if self.is_unsupervised():
            return 1 * (proba[:, 0] < self.proba_thr)
        else:
            return self.classes_[numpy.argmax(proba, axis=1)]

    def classifier_name(self):
        """
        Gets classifier name as string
        :return: the classifier name
        """
        clf_name = self.clf.classifier_name() if isinstance(self.clf, Classifier) else self.clf.__class__.__name__
        if clf_name == 'Pipeline':
            keys = list(self.clf.named_steps.keys())
            clf_name = str(keys) if len(keys) != 2 else str(keys[1]).upper()
        if self.weighted:
            return "ConfidenceBoosterWeighted(" + str(clf_name) + "-" + \
                   str(self.n_base) + "-" + str(self.conf_thr) + "-" + \
                   str(self.learning_rate) + "-" + str(self.sampling_ratio) + ")"
        else:
            return "ConfidenceBooster(" + str(clf_name) + "-" + \
                   str(self.n_base) + "-" + str(self.conf_thr) + "-" + \
                   str(self.learning_rate) + "-" + str(self.sampling_ratio) + ")"



import configparser
import os
import shutil
import time

import numpy as np


def load_config(file_config):
    """
    Method to load configuration parameters from input file
    :param file_config: name of the config file
    :return: array with 4 items: [dataset files,
                                    classifiers,
                                    label name,
                                    max number of rows (if correctly specified, NaN otherwise)]
    """
    config = configparser.RawConfigParser()
    if os.path.isfile(file_config):
        config.read(file_config)
        config_file = dict(config.items('CONFIGURATION'))
    
        # Processing classifiers
        s_classifiers = config_file['supervised_classifiers']
        if ',' in s_classifiers:
            s_classifiers = [x.strip() for x in s_classifiers.split(',')]
        else:
            s_classifiers = [s_classifiers]
        s_classifiers = [x for x in s_classifiers if x]
        u_classifiers = config_file['unsupervised_classifiers']
        if ',' in u_classifiers:
            u_classifiers = [x.strip() for x in u_classifiers.split(',')]
        else:
            u_classifiers = [u_classifiers]
        u_classifiers = [x for x in u_classifiers if x]

        # Folders
        d_folder = config_file['datasets_folder']
        if not d_folder.endswith("/"):
            d_folder = d_folder + "/"
        s_folder = config_file['sprout_scores_folder']
        if not s_folder.endswith("/"):
            s_folder = s_folder + "/"
    
        # Processing paths
        path_string = config_file['datasets']
        if ',' in path_string:
            path_string = [x.strip() for x in path_string.split(',')]
        else:
            path_string = [path_string]
        datasets_path = []
        for file_string in path_string:
            if file_string in ["MNIST", "DIGITS", "FASHION-MNIST"]:
                datasets_path.append(file_string)
            else:
                file_path = os.path.join(d_folder, file_string)
                if os.path.isdir(file_path):
                    datasets_path.extend([os.path.join(file_path, f) for f in os.listdir(file_path) if
                                          os.path.isfile(os.path.join(file_path, f))])
                else:
                    datasets_path.append(file_path)
    
        # Processing limit to rows
        lim_rows = config_file['limit_rows']
        if not lim_rows.isdigit():
            lim_rows = np.nan
        else:
            lim_rows = int(lim_rows)

        return datasets_path, d_folder, s_folder, s_classifiers, u_classifiers, config_file['label_tabular'], lim_rows
    
    else:
        # Config File does not exist
        return ["DIGITS"], ["RF"], "", "no"


def current_ms():
    """
    Reports the current time in milliseconds
    :return: long int
    """
    return round(time.time() * 1000)


def clean_name(file, prequel):
    """
    Method to get clean name of a file
    :param file: the original file path
    :return: the filename with no path and extension
    """
    if prequel in file:
        file = file.replace(prequel, "")
    if '.' in file:
        file = file.split('.')[0]
    if file.startswith("/"):
        file = file[1:]
    return file


def get_full_class_name(class_obj):
    return class_obj.__module__ + "." + class_obj.__qualname__


def clear_folder(folder_path):
    for filename in os.listdir(folder_path):
        file_path = os.path.join(folder_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))



import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
     name='confens',
     version='0.4',
     scripts=[],
     author="Tommaso Zoppi",
     author_email="tommaso.zoppi@unitn.it",
     description="Confidence Ensembles to Improve Classification",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/tommyippoz/confidence-ensembles",
     keywords=['machine learning', 'confidence', 'safety', 'ensemble'],
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
)
